/*
 * This is Tampa Ping-Pong convers/conversd derived from the wampes
 * convers package written by Dieter Deyke <deyke@mdddhd.fc.hp.com>
 * Modifications by Fred Baumgarten <dc6iq@insu1.etec.uni-karlsruhe.de>
 *
 * Modifications by Brian A. Lantz/KO4KS <brian@lantz.com> 
 * $Revision: 1.14 $$Date: 1995/06/16 14:47:52 $
 */

#define REV "$Revision: 1.14 $"

#include <sys/types.h>

#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#if defined(_AIX) && defined(_IBMR2)
#include <time.h>
#include <sys/select.h>
#endif

#if defined(__OS2__)
#include <time.h>
#include <types.h>
#include <netdb.h>
#include <nerrno.h>
#include <sys\ioctl.h>
#include <sys\select.h>
#include "pp_os2.h"
#define FD_SET_TYPE fd_set
#else
#include <sys/wait.h>
#include <unistd.h>
#include <utmp.h>
#endif

#ifdef LOGGING
#include <stdarg.h>
#include <netdb.h>
#include <netinet/in.h>
#if defined (mips)		/* rotten OS isn't it ? */
#include <sys/syslog.h>
#else
#include <syslog.h>
#endif
#include <arpa/inet.h>
#else
#ifdef linux
#include <netinet/in.h>          /* TODO - belongs to unix socket workaroud ! */
#endif
#endif

#include "conversd.h"
#include "convert.h"

#if defined(mips)      /* no headerfiles for the whole stuff ? */
extern int select __ARGS((int nfsd, fd_set *readfds, fd_set *writefds,
                  fd_set *exceptfds, struct timeval *timeout));
#endif

#if defined(hpux)
#define fd_set int	/* not sure about this... comments ? */
#endif

#if defined(mips) || defined(_AIX)     /* no headerfiles for the whole stuff ? */
extern void bzero __ARGS((char *b1, int length));
extern int socket __ARGS((int af, int type, int protocol));
extern int accept __ARGS((int s, struct sockaddr *addr, int *addrlen));
extern int connect __ARGS((int s, struct sockaddr *name, int namelen));
extern int putenv __ARGS((char *string));
extern void gethostname __ARGS((char *name, int namelen));
extern int setsockopt __ARGS((int s, int level, int optname, char *optval, int optlen));
extern int listen __ARGS((int s, int backlog));
extern int bind __ARGS((int s, struct sockaddr *name, int namelen));
#ifdef LOGGING
extern void closelog __ARGS((void));
extern void openlog __ARGS((__const char *, int, int));
extern void syslog __ARGS((int, __const char * __DOTS));
#endif
#endif

extern struct sockaddr *build_sockaddr __ARGS((const char *name, int *addrlen));

static char progfile[128];
static char conffile[128];
static char *convcmd;
static FD_SET_TYPE chkread;
static FD_SET_TYPE chkwrite;

char convtype[512];
long boottime;
long currtime;
char *myhostname = NULLCHAR;
char *mysysinfo = NULLCHAR;
char *myemailaddr = NULLCHAR;
char userfile[128];
char motdfile[128];
char restrictfile[128];
char accessfile[128];
char noaccessfile[128];
char helpfile[128];
char observerfile[128];
char issuefile[128];
char listener[128];
int NR_PERMLINKS = DEF_NR_PERMLINKS;
int secretnumber;
int amprnet;	/* if set to 0 it's an internet environment - should be site-specific */
int restrictedmode = 0; /* if set to 1, then all /name commands are treated
                           as /restricted commands, set to 2 they are treated
                           as /observer commands */
char myrev[80];
#ifdef oldAWAY
char myfeatures[] = "admpun";
#else
char myfeatures[] = "Aadmpun";
#endif
char OBSERVERSTR[] = "*** Sorry, you are only an observer :-)\n";
char OBSERVERSTR2[] = "*** That command is not available to observers....\n";
char OBSERVERSTR3[] = "*** Observers are not permitted to send messages....\n";
char OBSERVERID[] = "Observer only";

static int maxfd = -1;

struct connection *connections;
struct permlink **permarray = NULLPERMARRAY;
struct destination *destinations;
struct channel *channels;

struct extendedcmds *ecmds;

struct stats daily[60];
struct stats hourly[60];
time_t starttime;
time_t nextdailystats;
time_t nexthourlystats;
long hoursonline = 0L;
long daysonline = 0L;
int currentLOGlevel = LEV_LINKS;


#include "user.h"
#include "host.h"

static void process_input __ARGS((struct connection *cp));
static void read_configuration __ARGS((void));
#ifdef CHECKFILES
static void check_files_changed __ARGS((void));
#endif

extern void rip __ARGS((char *str));

struct cmdtable cmdtable[] = {
                                       
  {"/priv",		operator_command,	CM_USER},
  {"/quit",		bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
  {"?",			cmdsummary_command,	CM_USER | CM_OBSERVER},
  {"away",		away_command,		CM_USER},
  {"action",		me_command,		CM_USER},
  {"all",		all_command,		CM_USER},
  {"beep",		beep_command,		CM_USER | CM_OBSERVER},
  {"bell",		beep_command,		CM_USER | CM_OBSERVER},
  {"bye",		bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
  {"channel",		channel_command,	CM_USER | CM_OBSERVER},
  {"charset",		charset_command,	CM_USER | CM_OBSERVER},
  {"cq",		cq_command,		CM_USER},
  {"cstat",		cstat_command,		CM_UNKNOWN | CM_USER | CM_OBSERVER},
  {"cut",		cut_command,		CM_USER},
  {"destinations",	hosts_command,		CM_USER | CM_OBSERVER},
#ifdef SAMPLE
  {"ecmdtest",		ecmdtest_command,	CM_USER},
#endif
  {"exit",		bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
  {"exclude",		imsg_command,		CM_USER},
  {"help",		help_command,		CM_USER | CM_OBSERVER},
  {"halt",		restart_command,	CM_USER},
  {"hosts",		hosts_command,		CM_USER | CM_OBSERVER},
  {"invite",		invite_command,		CM_USER},
  {"imsg",		imsg_command,		CM_USER},
  {"iwrite",		imsg_command,		CM_USER},
  {"join",		channel_command,	CM_USER | CM_OBSERVER},
  {"links",		links_command,		CM_USER | CM_OBSERVER},
  {"last",		last_command,		CM_USER | CM_OBSERVER},
  {"leave",		leave_command,		CM_USER | CM_OBSERVER},
  {"list",		list_command,		CM_USER | CM_OBSERVER},
  {"msg",		msg_command,		CM_USER},
  {"me",		me_command,		CM_USER},
  {"mode",		mode_command,		CM_USER},
  {"name",		name_command,		CM_UNKNOWN},
  {"news",		news_command,		CM_USER | CM_OBSERVER},
  {"nickname",		nickname_command,	CM_USER},
  {"nonickname",	nonickname_command,	CM_USER},
  {"notify",		notify_command,		CM_USER},
  {"note",		personal_command,	CM_USER},
  {"observer",		observer_command,	CM_UNKNOWN},
  {"operator",		operator_command,	CM_USER},
  {"online",		who_command2,		CM_UNKNOWN},
  {"page",		invite_command,		CM_USER},
  {"personal",		personal_command,	CM_USER},
  {"prompt",		prompt_command,		CM_USER},
  {"quit",		bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
  {"query",		query_command,		CM_USER},
  {"quote",		quote_command,		CM_USER},
  {"realname",		realname_command,	CM_USER | CM_OBSERVER},
  {"restart",		restart_command,	CM_USER},
  {"restricted",	restricted_command,	CM_UNKNOWN},
  {"roll",		roll_command,		CM_USER},
  {"send",		msg_command,		CM_USER},
  {"smiley",		smiley_command,		CM_USER},
  {"stats",		stats_command,		CM_USER | CM_OBSERVER},
  {"sysinfo",		sysinfo_command,	CM_USER | CM_OBSERVER},
  {"sysop",		operator_command,	CM_USER},
#ifdef SAMPLE
  {"test1",		test1_command,		CM_USER},
#endif
  {"time",		time_command,		CM_USER},
  {"topic",		topic_command,		CM_USER},
  {"users",		who_command,		CM_USER},
  {"uptime",		uptime_command,		CM_USER | CM_OBSERVER},
  {"verbose",		verbose_command,	CM_USER},
  {"version",		version_command,	CM_USER | CM_OBSERVER},
  {"who",		who_command,		CM_USER | CM_OBSERVER},
  {"whois",		whois_command,		CM_USER | CM_OBSERVER},
  {"width",		width_command,		CM_USER | CM_OBSERVER},
  {"write",		msg_command,		CM_USER},
  
  {"\377\200away",	h_away_command,		CM_HOST},
  {"\377\200cmsg",	h_cmsg_command,		CM_HOST},
  {"\377\200dest",	h_dest_command,		CM_HOST},
  {"\377\200ecmd",	h_ecmd_command,		CM_HOST},
  {"\377\200help",	h_help_command,		CM_HOST},
  {"\377\200host",	h_host_command,		CM_UNKNOWN},
  {"\377\200invi",	h_invi_command,		CM_HOST},
  {"\377\200link",	h_link_command,		CM_HOST},
  {"\377\200loop",	h_loop_command,		CM_HOST},
  {"\377\200mode",	mode_command,		CM_HOST},
  {"\377\200obsv",	h_user_command,		CM_HOST},
  {"\377\200oper",	h_oper_command,		CM_HOST},
  {"\377\200ping",	h_ping_command,		CM_HOST},
  {"\377\200pong",	h_pong_command,		CM_HOST},
  {"\377\200rout",	h_rout_command,		CM_HOST},
  {"\377\200sysi",	h_sysi_command,		CM_HOST},
  {"\377\200topi",	h_topi_command,		CM_HOST},
  {"\377\200uadd",	h_uadd_command,		CM_HOST},
  {"\377\200udat",	h_udat_command,		CM_HOST},
  {"\377\200umsg",	h_umsg_command,		CM_HOST},
  {"\377\200user",	h_user_command,		CM_HOST},
  
  {NULL,		0,			0}
};

static char *getargp;

void setargp (char *str)
{
	getargp = str;
}


/*---------------------------------------------------------------------------*/

void appendstring(cp, string)
struct connection *cp;
const char *string;
{
  struct mbuf *bp, *p;
  char buffer[4096];
  const char *p_string;

  if (!*string) return;

  if ((cp->type == CT_USER) && (strncmp (string, "/\377\200", 3))) {
    strncpy(buffer, string, 4096);
    convert(ISO, cp->charset_out, buffer);
    p_string = buffer;
  } else p_string = string;

  bp = (struct mbuf *) malloc(sizeof(*bp) + strlen(p_string) + 1);
  bp->next = 0;
  bp->data = strcpy((char *) (bp + 1), p_string);

  if (cp->obuf) {
    for (p = cp->obuf; p->next; p = p->next) ;
    p->next = bp;
  } else {
    cp->obuf = bp;
    FD_SET(cp->fd, &chkwrite);
  }
}

/*---------------------------------------------------------------------------*/

void appendc(cp, n, ast)
struct connection *cp;
const int n;
const int ast;
{
  static char x[2];
  char c;

  c = cp->prompt[n];
  if (c == '\0') {
    if (ast) {
      appendstring(cp, "***\n");
    }
  } else {
    x[0] = c;
    x[1] = '\0';
    appendstring(cp, x);
  }
}

/*---------------------------------------------------------------------------*/

/* Find a permlink, given it's name, or NULL if no named permlink
 *    wkt@cs.adfa.oz.au
 */
struct permlink  *find_permlink(const char *name)
{
  int pl;
  struct permlink *p;

  for (pl = 0; pl < NR_PERMLINKS; pl++) {
        p = permarray[pl];
      if (p && !strcasecmp(p->name, name)) return(p);
  }
  return(NULL);
}


/*---------------------------------------------------------------------------*/

void appendprompt(cp, ast)
struct connection *cp;
const int ast;
{
  if (*cp->query)
    appendc(cp, 1, ast);
  else
    appendc(cp, 2, ast);
#ifdef POSIX
  fflush(NULL);
#endif
}

/*---------------------------------------------------------------------------*/

int queuelength(bp)
const struct mbuf *bp;
{
  int len;

  for (len = 0; bp; bp = bp->next)
    len += strlen(bp->data);
  return len;
}

/*---------------------------------------------------------------------------*/

void destroy_channel(number)
int number;
{
  struct channel *ch, *ch1;

  ch1 = NULLCHANNEL;
  for (ch = channels; ch; ch = ch->next) {
    if (ch->chan == number) {
      if (ch1) {
	ch1->next = ch->next;
      } else {
	channels = ch->next;
      }
      free(ch);
      break;
    }
    ch1 = ch;
  }
}

/*---------------------------------------------------------------------------*/

int ecmd_exists(cmdname)
char *cmdname;
{
struct extendedcmds *echk;

	echk = ecmds;
	while (echk && echk->name)	{
		if (!strcasecmp(echk->name, cmdname))
			return 1;
		echk = echk->next;
	}
	return 0;
}

/*---------------------------------------------------------------------------*/

void free_connection(cp)
struct connection *cp;
{
  struct mbuf *bp;
  struct permlink *p = NULLPERMLINK;
  int pl;

  for (pl = 0; pl < NR_PERMLINKS; pl++) {
    p = permarray[pl];
    if (p && (p->connection == cp)) {
      p->connection = NULLCONNECTION;
      break;
    }
  }
  if (cp->fd >= 0) {
#if defined(__OS2__)
    soclose(cp->fd);
#else
    close(cp->fd);
#endif
    FD_CLR(cp->fd, &chkread);
    FD_CLR(cp->fd, &chkwrite);
    if (cp->fd == maxfd)
      while (--maxfd >= 0)
	if (FD_ISSET(maxfd, &chkread)) break;
    cp->fd = -1;
  }
  while ((bp = cp->obuf) != NULL) {
    cp->obuf = bp->next;
    free(bp);
  }
  free(cp);
  if (pl < NR_PERMLINKS) {
    if (p && !p->permanent) {
      permarray[pl] = NULLPERMLINK;
    }
  }
}

/*---------------------------------------------------------------------------*/

void free_closed_connections()
{
  struct connection *cp, *p;
  struct permlink *l;
  int pl;
  
  for (p = NULLCONNECTION, cp = connections; cp; ) {
    l = NULLPERMLINK;
    for (pl = 0; pl < NR_PERMLINKS; pl++) {
      if ((permarray[pl]) && (permarray[pl]->connection == cp)) {
	l = permarray[pl];
	break;
      }
    }
    if (l) {
      if (cp->type == CT_CLOSED ||
	  (cp->type == CT_UNKNOWN && cp->time + l->waittime - 5 < currtime)) {
	if (p) {
	  p->next = cp->next;
	  free_connection(cp);
	  cp = p->next;
	} else {
	  connections = cp->next;
	  free_connection(cp);
	  cp = connections;
	}
      } else {
	p = cp;
	cp = cp->next;
      }
    } else {
      if (cp->type == CT_CLOSED) {
	if (p) {
	  p->next = cp->next;
	  free_connection(cp);
	  cp = p->next;
	} else {
	  connections = cp->next;
	  free_connection(cp);
	  cp = connections;
	}
      } else {
	p = cp;
	cp = cp->next;
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

char *getarg(line, all)
char *line;
int all;
{
  char *arg;
  int c;

  if (line) getargp = line;
  while (isspace(uchar(*getargp))) getargp++;
  if (all) return getargp;
  arg = getargp;
  while (*getargp && !isspace(uchar(*getargp))) {
    c = tolower(uchar(*getargp));
    *getargp++ = c;
  }
  if (*getargp) *getargp++ = '\0';
  return arg;
}

/*---------------------------------------------------------------------------*/

char *getargcs(line, all)
char *line;
int all;
{
  char *arg;

  if (line) getargp = line;
  while (isspace(uchar(*getargp))) getargp++;
  if (all) return getargp;
  arg = getargp;
  while (*getargp && !isspace(uchar(*getargp))) {
    getargp++;
  }
  if (*getargp) *getargp++ = '\0';
  return arg;
}

/*---------------------------------------------------------------------------*/

void getTXname(cp, buf)
struct connection *cp;
char *buf;
{
	if (!strcasecmp (cp->nickname, cp->name))
		strcpy (buf, cp->name);
	else
		sprintf (buf, "%s:%s", cp->name, cp->nickname);
}


/*---------------------------------------------------------------------------*/

char *formatline(prefix, text, linelen)
char *prefix, *text;
int linelen;
{
  char *f, *t, *x;
  int prefixlen, l, lw;
  static char buf[2048];

  linelen--;
  prefixlen = strlen(prefix);
  for (f = prefix, t = buf; *f; *t++ = *f++) ;
  l = t - buf;
  f = text;

  for (; ; ) {
    while (isspace(uchar(*f))) f++;
    if (!*f) {
      *t++ = '\n';
      *t = '\0';
      return buf;
    }
    for (x = f; *x && !isspace(uchar(*x)); x++) ;
    lw = x - f;
    if (l > prefixlen && l + 1 + lw > linelen) {
      *t++ = '\n';
      l = 0;
    }
    do {
      *t++ = ' ';
      l++;
    } while (l < prefixlen);
    while (lw--) {
      *t++ = *f++;
      l++;
    }
  }
}

/*---------------------------------------------------------------------------*/

char *ts(gmt)
long gmt;
{
  static char buffer[80];
  static char monthnames[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
  struct tm *tm;

  tm = localtime(&gmt);
  if (gmt + 24 * 60 * 60 > currtime)
    sprintf(buffer, " %2d:%02d", tm->tm_hour, tm->tm_min);
  else
    sprintf(buffer, "%-3.3s %2d", monthnames + 3 * tm->tm_mon, tm->tm_mday);
  return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts2(gmt)
long gmt;
{
  static char buffer[80];
  static char monthnames[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
  struct tm *tm;

  tm = localtime(&gmt);
  if (tm)	{
	  if (gmt + 24 * 60 * 60 > currtime)
	    sprintf(buffer, "%2d:%02d", tm->tm_hour, tm->tm_min);
	  else
	    sprintf(buffer, "%-3.3s %2d", monthnames + 3 * tm->tm_mon, tm->tm_mday);
  } else
	strcpy (buffer, "?????");
  return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts3(seconds,buffer)
long seconds;
char *buffer;
{
  if (seconds < 100) {
    sprintf(buffer, "%lds", seconds);
  } else if (seconds <6000) {
    sprintf(buffer, "%ldm", seconds/60);
  } else if (seconds <360000) {
    sprintf(buffer, "%ldh", seconds/3600);
  } else {
    sprintf(buffer, "%ldd", seconds/86400);
  }
  return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts4(seconds)
long seconds;
{
  int days, hours, minutes;
  static char buffer[2048];
  char buf[128];

  days = seconds / 86400;
  seconds -= days * 86400;
  hours = seconds / 3600;
  seconds -= hours * 3600;
  minutes = seconds / 60;
  seconds -= minutes * 60;
  buffer[0] = '\0';
  if (days) {
    sprintf(buf, "%d day%s, ", days, (days == 1) ? "" : "s");
    strncpy(buffer, buf, 2048);
  }
  if (days+hours) {
    sprintf(buf, "%d hour%s, ", hours, (hours == 1) ? "" : "s");
    strcat(buffer, buf);
  }
  if (days+hours+minutes) {
    sprintf(buf, "%d minute%s, ", minutes, (minutes == 1) ? "" : "s");
    strcat(buffer, buf);
  }
  sprintf(buf, "%ld second%s.", seconds, (seconds == 1) ? "" : "s");
  strcat(buffer, buf);
  return buffer;
}

/*---------------------------------------------------------------------------*/

struct connection *alloc_connection(fd, linuxbug)
int fd;
int linuxbug;
{
  int flags;
  struct connection *cp;
#if defined(__OS2__)
  int mask = 1;
#endif

#if defined(__OS2__)
  if (ioctl(fd, FIONBIO, (unsigned char *)&mask, sizeof(mask)) == -1) {
    soclose(fd);
    return(0);
  }
#else
#if defined(linux)
  if (linuxbug == 1) {
#endif
  if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
    close(fd);
    return 0;
  }
  if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
    close(fd);
    return 0;
#if defined(linux)
  }
#endif
  }
#endif
  cp = (struct connection *) calloc(1, sizeof(*cp));
  cp->fd = fd;
  cp->time = currtime;
  cp->verbose = 0;
  cp->prompt[0] = 0;
  cp->prompt[1] = 0;
  cp->prompt[2] = 0;
  cp->prompt[3] = 0;
  cp->features = 0;
  *cp->rev = '\0';
  *cp->away = '\0';
  cp->chan_list = NULLCLIST;
  cp->amprnet = amprnet;
  cp->oldaway = 1;
  cp->atime = currtime;
  cp->mtime = 0L;
  *cp->pers = '\0';
  cp->width = 80;
  cp->observer = 0;
  cp->next = connections;
  connections = cp;
  if (maxfd < fd) maxfd = fd;
  FD_SET(fd, &chkread);
  return cp;
}

/*---------------------------------------------------------------------------*/

void accept_connect_request(flisten)
int flisten;
{
  FILE *fdi;
  char buffer[2048];
  int addrlen;
  int fd, flags;
  struct sockaddr addr;
  struct connection *cp;
#ifdef LOGGING
  struct sockaddr_in *sin = (struct sockaddr_in *)&addr;
  struct hostent *hp;
  char hostname[128];
  int allow;
  extern int hosts_ctl __ARGS((char *daemon, char *name, char *addr, char *user));
#endif
#if defined(__OS2__)
  int mask = 1;
#endif

  addrlen = sizeof(addr);
  if ((fd = accept(flisten, &addr, &addrlen)) < 0) return;
#if defined (__OS2__)
  if (ioctl(fd, FIONBIO, (unsigned char *)&mask, sizeof(mask)) == -1 ) {
#else
  if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
    close(fd);
    return;
  }
  if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
#endif
    logit (LEV_LINKS, "fcntl failed connection request, reason %d", errno);
#if defined(__OS2__)
    soclose(fd);
#else
    close(fd);
#endif
    return;
  }

#ifdef LOGGING
  if(sin->sin_family == AF_INET){
    if ((hp = gethostbyaddr((char *) &sin->sin_addr, sizeof(sin->sin_addr), AF_INET)) != NULL) {
      strncpy(hostname, hp->h_name, 128);
      if (strcmp(hostname, "localhost")) {
        logit(LEV_LINKS, "connect from %s", hostname);
      }
    } else {
      logit(LEV_LINKS, "connect from %s", inet_ntoa(sin->sin_addr));
    }
    allow = hosts_ctl(convtype, "", inet_ntoa(sin->sin_addr), "");
    if (!allow){
      logit(LEV_LINKS, "refused connect from %s",  inet_ntoa(sin->sin_addr));
#if defined (__OS2__)
      soclose(fd);
#else
      close(fd);
#endif
      return;
    }
  }
#endif
  cp = alloc_connection(fd, 0);
  if (!access(issuefile, R_OK)) {
    fdi = fopen(issuefile, "r");
    buffer[0] = '*';
    buffer[1] = ' ';
    while (!feof(fdi)) {
      fgets(buffer+2, 2045, fdi);
      if (!feof(fdi)) {
        appendstring(cp, buffer);
      }
    }
    fclose(fdi);
  }
}

/*---------------------------------------------------------------------------*/

int count_user(channel)
int channel;
{
  struct connection *p;
  struct clist *cl;
  int n = 0;

  for (p = connections; p; p = p->next) {
    if (p->type == CT_USER) {
      if (p->via && (channel == -1 || p->channel == channel)) {
	n++;
      } else {
	for (cl = p->chan_list; cl; cl = cl->next) {
	  if (channel == -1 || cl->channel == channel) {
	    n++;
	    break;
	  }
	}
      }
    }
  }
  return n;
}

/*---------------------------------------------------------------------------*/

int count_topics(void)
{
int n = 0;
register struct channel *ch;

	for (ch = channels; ch; ch = ch->next)
		if (!(ch->flags & (M_CHAN_S || M_CHAN_I)) && *ch->topic)
			n++;
	return n;
}

/*---------------------------------------------------------------------------*/

int isonchannel(cp, user, host)
struct connection *cp;
char *user;
char *host;
{
struct connection *p;
struct clist *cl, *cl1;

	for (p = connections; p; p = p->next) {
		if ((p->type == CT_USER) && (!p->via) && (!strcasecmp(p->name, user))) {
			for (cl1 = cp->chan_list; cl1; cl1 = cl1->next) {
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cl->channel == cl1->channel)
						return 1;
				}
			}
		}
	}
	return 0;
}

/*---------------------------------------------------------------------------*/
 
void clear_locks()
{
  struct connection *p;

  for (p = connections; p; p = p->next) p->locked = 0;
}

/*---------------------------------------------------------------------------*/

void notify_destinations (dest, host, rev, state)
struct destination *dest;
const char *host, *rev;
int state;
{
register struct connection *p;
char buffer[2048];
char revstr[40];

	if (dest && ((state && dest->connected) || (!state && !dest->connected)))
		return;
	clear_locks();
	if (rev && *rev)
		sprintf (revstr, "(%.8s)", rev);
	else
		revstr[0] = 0;
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER && !p->via && (p->operator || p->channelop)) {
			sprintf(buffer, "*** (%s) Server %sing the network: %s %s\n", ts2(currtime), (state) ? "join" : "leav", host, revstr);
			appendstring(p, buffer);
			appendprompt(p, 0);
		}
	}
	if (dest)
		dest->connected = state;
}

/*---------------------------------------------------------------------------*/

#ifdef oldAWAY
void send_awaymsg(fromname, hostname, channel, time, text)
int channel;
#else
void send_awaymsg(fromname, hostname, time, text)
#endif
char *fromname, *hostname;
long time;
char *text;
{
  char buffer[2048];
#ifdef oldAWAY
  register struct clist *cl;
#endif
  register struct connection *p;

  for (p = connections; p; p = p->next) {
    if (p->type == CT_HOST) {
      if (!p->locked) {
#ifdef oldAWAY
	sprintf(buffer, "/\377\200AWAY %s %s %d %ld %s\n", fromname, hostname, channel, time, text);
#else
        if (p->oldaway) {
	  sprintf(buffer, "/\377\200AWAY %s %s 69 %ld %s\n", fromname, hostname, time, text);
        } else {
	  sprintf(buffer, "/\377\200AWAY %s %s %ld %s\n", fromname, hostname, time, text);
        }
#endif
	logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	appendstring(p, buffer);
	p->locked = 1;
      }
    } else {
      if (!p->locked && !p->via) {
#ifdef oldAWAY
	for (cl = p->chan_list; cl; cl = cl->next) {
	  if (cl->channel == channel) break;
	}
	if (cl) {
#else
	if (isonchannel(p, fromname, hostname)) {
#endif
	  appendc(p, 0, 0);
	  appendc(p, 3, 0);
	  if (*text != '\0')
	    sprintf(buffer, "*** (%s) %s@%s has gone away:\n    %s\n", ts2(currtime), fromname, hostname, text);
	  else
	    sprintf(buffer, "*** (%s) %s@%s is back again.\n", ts2(currtime), fromname, hostname);
	  appendstring(p, buffer);
	  appendprompt(p, 0);
#ifdef oldAWAY
	}
#endif
	p->locked = 1;
#ifndef oldAWAY
       }
#endif
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_mode(ch)
struct channel *ch;
{
  register struct connection *p;
  char buffer[2048];
  char flags[16];

  flags[0] = '\0';
  if (ch->flags & M_CHAN_S) {
    strcat(flags, "S");
  }
  if (ch->flags & M_CHAN_P) {
    strcat(flags, "P");
  }
  if (ch->flags & M_CHAN_T) {
    strcat(flags, "T");
  }
  if (ch->flags & M_CHAN_I) {
    strcat(flags, "I");
  }
  if (ch->flags & M_CHAN_M) {
    strcat(flags, "M");
  }
  if (ch->flags & M_CHAN_L) {
    strcat(flags, "L");
  }
  for (p = connections; p; p = p->next) {
    if (!p->locked && (p->type == CT_HOST)) {
      sprintf(buffer, "/\377\200MODE %d -sptiml+%s\n", ch->chan, flags);
      logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
      appendstring(p, buffer);
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_opermsg(toname, hostname, fromname, channel)
char *toname, *hostname, *fromname;
int channel;
{
  char buffer[2048];
  register struct connection *p;

  for (p = connections; p; p = p->next) {
    if (!p->locked) {
      if (p->type == CT_HOST) {
	sprintf(buffer, "/\377\200OPER %s %d %s\n", fromname, channel, toname);
	logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	appendstring(p, buffer);
	p->locked = 1;
      } else {
	if (p->type == CT_USER) {
	  if (!p->via && (channel != -1)) {
	    if ((!strcasecmp(p->name,toname) || !strcasecmp(p->nickname,toname)) && (p->channel == channel)) {
	      appendc(p, 0, 0);
	      appendc(p, 3, 0);
	      sprintf(buffer, "*** (%s) %s made you a channel operator for channel %d\n",
		      ts2(currtime), fromname, channel);
	      appendstring(p, buffer);
	      appendprompt(p, 0);
	      p->locked = 1;
	    } else {
	      if (p->verbose) {
		appendc(p, 0, 0);
		appendc(p, 3, 0);
		sprintf(buffer, "*** (%s) %s@%s is now a channel operator for channel %d\n",
			ts2(currtime), toname, hostname, channel);
		appendstring(p, buffer);
		appendprompt(p, 0);
		p->locked = 1;
	      }
	    }
	  }
	}
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_persmsg(fromname, hostname, channel, text, time)
char *fromname, *hostname;
int channel;
char *text;
long time;
{
  char buffer[2048];
  register struct connection *p;
  struct channel *ch;
  char chan[128];
  int mychannel;
  struct clist *cl;

  for (ch = channels; ch; ch = ch->next) {
    if (ch->chan == channel) break;
  }
  sprintf(chan, "channel %d", channel);
  if (ch) {
    if (ch->flags & M_CHAN_S) sprintf(chan, "secret channel %d", channel);
    if (ch->flags & M_CHAN_I) sprintf(chan, "invisible channel %d", channel);
  }
  for (p = connections; p; p = p->next) {
    if (p->type == CT_HOST) {
      if (!p->locked) {
	if (p->amprnet) {
	  sprintf(buffer, "/\377\200USER %s %s %ld %d %d %s\n", fromname, hostname, time, channel, channel, text);
	} else {
	  sprintf(buffer, "/\377\200UDAT %s %s %s\n", fromname, hostname, text);
	}
	logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	appendstring(p, buffer);
	p->locked = 1;
      }
    } else {
      mychannel = -1;
      for (cl = p->chan_list; cl; cl = cl->next) {
	if (cl->channel == channel) {
	  mychannel = channel;
	  break;
	}
      }
      if ((p->type == CT_USER) && !p->locked && !p->via &&
	  (p->verbose || (mychannel == channel))) {
	appendc(p, 0, 0);
	appendc(p, 3, 0);
	if ((*text != '\0') && (strcmp(text, "@"))) {
	  sprintf(buffer, "*** (%s) %s@%s on %s set personal text:\n    %s\n", ts2(currtime), fromname, hostname, chan, text);
	} else {
	  sprintf(buffer, "*** (%s) %s@%s on %s removed personal text.\n", ts2(currtime), fromname, hostname, chan);
	}
	appendstring(p, buffer);
	appendprompt(p, 0);
      }
      p->locked = 1;
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_topic(fromname, hostname, time, channel, text)
char *fromname, *hostname;
long time;
int channel;
char *text;
{
  char buffer[2048];
  register struct connection *p;
  register struct channel *ch;
  char chan[128];
  int mychannel, flags = 0;
  struct clist *cl;

  for (ch = channels; ch; ch = ch->next)
    if (ch->chan == channel) break;
  if (ch) {
    sprintf(chan, "channel %d", channel);
    flags = ch->flags;
    if (flags & M_CHAN_S) sprintf(chan, "secret channel %d", channel);
    if (flags & M_CHAN_I) sprintf(chan, "invisible channel %d", channel);
    if (ch->time < time) {
      strncpy(ch->topic, text, TOPICSIZE);
      ch->time = time;
      for (p = connections; p; p = p->next) {
        if (p->type == CT_HOST) {
	  if (!p->locked) {
	    sprintf(buffer, "/\377\200TOPI %s %s %ld %d %s\n", fromname, hostname, time, channel, text);
	    logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	    appendstring(p, buffer);
	    p->locked = 1;
	  }
        } else {
	  mychannel = -1;
	  for (cl = p->chan_list; cl; cl = cl->next) {
	    if (cl->channel == channel) {
	      mychannel = channel;
	      break;
	    }
	  }
	  if ((p->type == CT_USER) && !p->locked && !p->via && 
	        ((!(ch->flags & M_CHAN_I) && p->verbose) || (mychannel==channel) )) {
	    appendc(p, 0, 0);
	    appendc(p, 3, 0);
	    if (*text != '\0')
	      sprintf(buffer, "*** (%s) %s@%s on %s set channel topic:\n    %s\n", ts2(currtime), fromname, hostname, chan, text);
	    else
	      sprintf(buffer, "*** (%s) %s@%s on %s removed channel topic.\n", ts2(currtime), fromname, hostname, chan);
	    appendstring(p, buffer);
	    appendprompt(p, 0);
	  }
	  p->locked = 1;
        }
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

int get_hash(name)
char *name;
{
  int i, hash;
  
  hash = 0;
  for (i = 0; i < strlen(name); i ++) {
    hash = ((hash << 1) + name[i]) & 0xff;
  }
  return hash;
}
  

/*---------------------------------------------------------------------------*/

void log_user_change(name, host, oldchannel, newchannel)
char *name, *host;
int oldchannel, newchannel;
{
#if WANT_LOG
#ifndef oldLOG
  char fname[128];
  sprintf(fname, "%s%02x", userfile, get_hash(name));
#endif
  if (oldchannel < 0 || newchannel < 0) {
    FILE *f;
    long found;
    char *p, *q;
    char xname[512];

#ifdef oldLOG
    f = fopen(userfile,"r+");
#else
    f = fopen(fname, "r+");
#endif
    if (f == NULL) {
#ifdef oldLOG
      f = fopen(userfile,"w+");
#else
      f = fopen(fname, "w+");
#endif
      if (f == NULL)
	return;
    }
    
    q = xname;
    for (p = name; isalnum(*p); p++) *q++ = *p;
    *q++ = '@';
    *q = '\0';
    strcat(xname, host);

    /* zuerst die passende zeile suchen */
    found = -1l;
    while (!feof(f)) {
      char buffer[512];
      found = ftell(f);
      fgets(buffer, 512, f);
      if (!strncmp(buffer,xname,strlen(xname)) 	/*  && 
	  ((newchannel < 0 && strstr(buffer," +")) ||
	   (oldchannel < 0 && strstr(buffer," -")))	*/ )
	break;
    }

    rewind(f);
    if (found >= 0l) {
      fseek(f, found, SEEK_SET);
    }
    else {
      fclose(f);
      f = fopen(userfile,"a");
    }

    if (f != NULL) {
      struct tm *tm;

      tm = localtime(&currtime);
      fprintf(f,"%-20s %2d.%02d.%02d %2d:%02d %c\n", 
	      xname, tm->tm_mday,tm->tm_mon+1,tm->tm_year,tm->tm_hour,tm->tm_min,
	      (oldchannel < 0) ? '+' : '-');
    }
    fclose(f);
  }
#endif
}

/*---------------------------------------------------------------------------*/

void send_user_change_msg(name, host, oldchannel, newchannel, pers, time, observer)
char *name, *host;
int oldchannel, newchannel;
char *pers;
long time;
int observer;
{
  char buffer[2048];
  struct connection *p;
  char oldchan[128], newchan[128];
  struct channel *ch;
  int oldflags = 0, newflags = 0;
  int mychannel;
  struct clist *cl;

  sprintf(newchan, "channel %d", newchannel);
  sprintf(oldchan, "channel %d", oldchannel);
  for (ch = channels; ch; ch = ch->next)
    if (ch->chan == newchannel) break;
  if (ch) {
    newflags = ch->flags;
    if (newflags & M_CHAN_S) {
      sprintf(newchan, "secret channel %d", newchannel);
    }
    if (newflags & M_CHAN_I) {
      sprintf(newchan, "invisible channel %d", newchannel);
    }
  }
  for (ch = channels; ch; ch = ch->next)
    if (ch->chan == oldchannel) break;
  if (ch) {
    oldflags = ch->flags;
    if (oldflags & M_CHAN_S) {
      sprintf(oldchan, "secret channel %d", oldchannel);
    }
    if (oldflags & M_CHAN_I) {
      sprintf(oldchan, "invisible channel %d", oldchannel);
    }
  }
  log_user_change(name, host, oldchannel,newchannel);
  for (p = connections; p; p = p->next) {
    mychannel = -1;
    if (p->type == CT_USER && !p->via && !p->locked) {
      for (cl = p->chan_list; cl; cl = cl->next) {
	if (cl->channel == oldchannel) {
	  mychannel = oldchannel;
	  break;
	}
      }
      if ((newchannel == oldchannel) && (newchannel != -1)) {
	if ((p->verbose && !(newflags & M_CHAN_I)) || (mychannel == newchannel)) {
	  appendc(p, 0, 0);
	  appendc(p, 3, 0);
	  if ((*pers != '\0') && (strcmp(pers, "@")))
            sprintf(buffer, "*** (%s) %s@%s on %s set personal text:\n    %s\n", ts2(currtime), name, host, newchan, pers);
          else
            sprintf(buffer, "*** (%s) %s@%s on %s removed personal text.\n", ts2(currtime), name, host, newchan);
          appendstring(p, buffer);
          appendprompt(p, 0);
          p->locked = 1;
	}
      } else {
	if (oldchannel >= 0) {
	  if ((!(oldflags & M_CHAN_I) && p->verbose) || (mychannel == oldchannel)) {
	    appendc(p, 0, 0);
	    appendc(p, 3, 0);
	    if (pers && (*pers != '\0') && strlen(pers) < 2) {
	      sprintf(buffer, "*** (%s) %s@%s left %s.\n", ts2(currtime), name, host, oldchan);
	    } else {
	      sprintf(buffer, "*** (%s) %s@%s left %s (%s).\n", ts2(currtime), name, host, oldchan, pers);
	    }
	    logit (LEV_USERS, "%s@%s left %s", name, host, oldchannel);
	    appendstring(p, buffer);
	    appendprompt(p, 0);
	    p->locked = 1;
	  }
	}
	for (cl = p->chan_list; cl; cl=cl->next) {
	  if (cl->channel == newchannel) {
	    break;
	  }
	}
	if (newchannel >= 0) {
	  sprintf(buffer, " %s ", name);
	  if ((cl) || (!(newflags & M_CHAN_I) && (p->verbose || strstr(p->notify, buffer)))) {
	    appendc(p, 0, 0);
	    appendc(p, 3, 0);
	    sprintf(buffer, "*** (%s) %s@%s joined %s\n", ts2(currtime), name, host, newchan);
	    appendstring(p, buffer);
	    logit (LEV_USERS, "%s@%s joined %s", name, host, newchan);
	    appendprompt(p, 0);
	    p->locked = 1;
	  }
	}
      }
    }
    if (p->type == CT_HOST && !p->locked) {
      if (time == currtime) {
	time++;
      }
      if (p->amprnet) {
	sprintf(buffer, "/\377\200%s %s %s %ld %d %d %s\n", (observer) ? "OBSV" : "USER", name, host, time, oldchannel, newchannel, pers);
      } else {
	sprintf(buffer, "/\377\200%s %s %s %ld %d %d\n", (observer) ? "OBSV" : "USER", name, host, time, oldchannel, newchannel);
	if (newchannel != -1) {
	  if ((oldchannel != -1) || ((pers) && (*pers) && (*pers != '@'))) {
	    if (!observer)	{
		logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	    	appendstring(p, buffer);
	    	sprintf(buffer, "/\377\200UDAT %s %s %s\n", name, host, pers);
	    }
	  }
	}
      }
      logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
      appendstring(p, buffer);
      p->locked = 1;
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_msg_to_local_user(fromname, toname, text)
char *fromname, *toname, *text;
{

  char buffer[2048];
  struct connection *p;
  char buf[(NAMESIZE * 2) + 2];

  for (p = connections; p; p = p->next)
    if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)) && !p->locked)
      if (!p->via) {
        if (!p->locked) {
          appendc(p, 0, 0);
          appendc(p, 3, 0);
	  if (strcmp(fromname, "conversd")) {
            sprintf(buffer, "<*%s*>:", fromname);
            appendstring(p, formatline(buffer, text, p->width));
            if (p->away[0] != '\0' && text[0] != '*') {
	      getTXname (p, buf);
              sprintf(buffer,"*** (%s) %s is away: %s", ts2(currtime), buf, p->away);
              send_msg_to_local_user("conversd",fromname,buffer);
            }
          } else {
            appendstring(p, text);
            appendstring(p, "\n");
          }
          appendprompt(p, 0);
          p->locked = 1;
        }
      }
}

/*---------------------------------------------------------------------------*/

void send_msg_to_user(fromname, toname, text)
char *fromname, *toname, *text;
{

  char buffer[2048];
  struct connection *p;
  char buf[(NAMESIZE * 2) + 2];

  for (p = connections; p; p = p->next)
    if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)) && !p->locked)
      if (p->via) {
	if (!p->via->locked) {
	  sprintf(buffer, "/\377\200UMSG %s %s %s\n", fromname, toname, text);
	  logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	  appendstring(p->via, buffer);
	  p->via->locked = 1;
	  p->locked = 1;
	}
      } else {
	if (!p->locked) {
          appendc(p, 0, 0);
	  appendc(p, 3, 0);
	  if (strcmp(fromname, "conversd")) {
	    sprintf(buffer, "<*%s*>:", fromname);
	    appendstring(p, formatline(buffer, text, p->width));
            if (p->away[0] != '\0' && text[0] != '*') {
	      getTXname(p, buf);
              sprintf(buffer,"*** (%s) %s is away: %s", ts2(currtime), buf, p->away);
              send_msg_to_local_user("conversd",fromname,buffer);
            }
 	  } else {
	    appendstring(p, text);
	    appendstring(p, "\n");
	  }
          appendprompt(p, 0);
	  p->locked = 1;
	}
      }
}

/*---------------------------------------------------------------------------*/

void send_msg_to_channel(fromname, channel, text)
char *fromname;
int channel;
char *text;
{
  char buffer[2048];
  struct connection *p;
  struct channel *ch;
  char addchan[16];
  struct clist *cl;
  int printit;

  for (ch = channels; ch; ch = ch->next) {
    if (ch->chan == channel) break;
  }
  for (p = connections; p; p = p->next) {
    if (p->type == CT_USER && (!strcasecmp(p->name, fromname) || !strcasecmp(p->nickname, fromname))) {
      p->mtime = currtime;
    }
  }
  for (p = connections; p; p = p->next) {
    if (p->type == CT_USER) {
      printit = 0;
      if (p->channel == channel) {
	printit = 1;
	addchan[0] = '\0';
      } else {
	for (cl = p->chan_list; cl ; cl = cl->next) {
	  if (cl->channel == channel) {
	    printit = 1;
	    sprintf(addchan, "%d:", channel);
	  }
	}
      }
      if (printit) {
	if (p->via) {
	  if (!p->via->locked && !(ch->flags & M_CHAN_L)) {
	    sprintf(buffer, "/\377\200CMSG %s %d %s\n", fromname, channel, text);
	    logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	    appendstring(p->via, buffer);
	    p->via->locked = 1;
	  }
	} else {
	  if (!p->locked) {
	    appendc(p, 0, 0);
	    appendc(p, 3, 0);
	    if (strcmp(fromname, "conversd")) {
	      sprintf(buffer, "<%s%s>:", addchan, fromname);
	      appendstring(p, formatline(buffer, text, p->width));
	    } else {
	      appendstring(p, text);
	      appendstring(p, "\n");
	    }
	    appendprompt(p, 0);
	    p->locked = 1;
	  }
	}
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

void send_invite_msg(fromname, toname, channel)
char *fromname, *toname;
int channel;
{

  static char invitetext[] = "\n\007\007*** (%s) Message from %s...\nPlease join %s channel %d.\n\007\007\n";

  static char responsetext[] = "*** (%s) %s Invitation sent to %s @ %s.";

  char buffer[2048];
  char convtype2[2048];
  int fdtty;
  int fdut;
  struct connection *p;
  struct stat stbuf;
#if !defined(__OS2__)
  struct utmp utmpbuf;
#endif

  strncpy(convtype2, convtype, 2048);
  convtype2[strlen(convtype2) - 1] = '\0';
  for (p = connections; p; p = p->next)		{
    if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname))) {
      if (p->channel == channel) {
	clear_locks();
	sprintf(buffer, "*** (%s) User %s is already on this channel.", ts2(currtime), toname);
	send_msg_to_user("conversd", fromname, buffer);
	return;
      }
      if (!p->via && !p->locked) {
	sprintf(buffer, invitetext, ts2(currtime), fromname, convtype2, channel);
	p->invitation_channel = channel;
	appendstring(p, buffer);
	clear_locks();
	sprintf(buffer, responsetext, ts2(currtime), convtype2, toname, myhostname);
	send_msg_to_user("conversd", fromname, buffer);
	return;
      }
      if (p->via && !p->via->locked) {
	sprintf(buffer, "/\377\200INVI %s %s %d\n", fromname, toname, channel);
	logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	appendstring(p->via, buffer);
	return;
      }
    }
  }

#if !defined(__OS2__)
    if ((fdut = open("/etc/utmp", O_RDONLY, 0644)) >= 0) {
      int found = 0;
      while (read(fdut, (char *)&utmpbuf, sizeof(utmpbuf)) == sizeof(utmpbuf)) {
	if (
#ifdef USER_PROCESS
	    utmpbuf.ut_type == USER_PROCESS &&
#endif
	    !strncmp(utmpbuf.ut_name, toname, sizeof(utmpbuf.ut_name))
	   ) {
          found++;
	  strcpy(buffer, "/dev/");
	  strncat(buffer, utmpbuf.ut_line, sizeof(utmpbuf.ut_line));
	  if (stat(buffer, &stbuf)) continue;
	  if (!(stbuf.st_mode & 2)) continue;
	  if ((fdtty = open(buffer, O_WRONLY | O_NOCTTY, 0644)) < 0) continue;
	  sprintf(buffer, invitetext, ts2(currtime), fromname, convtype2, channel);
	  if (!fork()) {
	    write(fdtty, buffer, strlen(buffer));
	    exit(0);
	  }
	  close(fdtty);
        }
      }
      if (found)	{
	  clear_locks();
	  sprintf(buffer, responsetext, ts2(currtime), convtype2, toname, myhostname);
	  send_msg_to_user("conversd", fromname, buffer);
	  close(fdut);
	  return;
	}
      close(fdut);
    }
#endif

  for (p = connections; p; p = p->next) {
    if (p->type == CT_HOST && !p->locked) {
      sprintf(buffer, "/\377\200INVI %s %s %d\n", fromname, toname, channel);
      logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
      appendstring(p, buffer);
    }
  }
}


/*---------------------------------------------------------------------------*/

void update_destinations(p, name, rtt, rev)
struct permlink *p;
char *name;
long rtt;
char *rev;
{
  struct permlink *l;
  struct destination *d, *d1;
  struct connection *cp;
  int pl;
  char buffer[2048];
  
  for (d = destinations; d; d = d->next)
    if (!strcasecmp(d->name, name)) break;
  
  if (!d) {
    d = (struct destination *)malloc(sizeof(struct destination));
    if (d) {
      d->last_sent_rtt = 0;
      d->link = p;
      d->rtt = 0;
      strncpy(d->name, name, HOSTNAMESIZE);
      strncpy(d->rev, rev, NAMESIZE);
      if (!destinations || (! destinations->name) || (strcasecmp(destinations->name, name) > 0)) {
	d->next = destinations;
	destinations = d;
      } else {
	d1 = destinations;
	while (d1->next) {
	  if (strcasecmp(d1->next->name, name) > 0) {
	    d->next = d1->next;
	    d1->next = d;
	    break;
	  }
	  d1 = d1->next;
	}
	if (!d1->next) {
	  d->next = d1->next;
	  d1->next = d;
	}
      }
    }
  } else {
    strncpy(d->rev, rev, NAMESIZE);
    d->link = p;
  }
      
  if (!d->rtt && rtt)	{
	for (cp = connections; cp; cp = cp->next) {
		if (!strcasecmp(cp->name, name))
			break;
	}
	if (cp == NULLCONNECTION)
		notify_destinations(d, name, rev, 1);
  }
  d->rtt = rtt;
  if (abs(rtt - d->last_sent_rtt) > (d->last_sent_rtt / 8)) {
    for (pl = 0; pl < NR_PERMLINKS; pl++) {
      l = permarray[pl];
      if (l && (l->connection) && (l->connection->type == CT_HOST)) {
	if (rtt) {
	  sprintf(buffer, "/\377\200DEST %s %ld %s\n", name, rtt+(l->txtime+l->rxtime)/2L, rev);
	} else {
	  sprintf(buffer, "/\377\200DEST %s 0\n", name);
	}
	if (strcasecmp(d->link->name, l->name)) {
	  logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
	  appendstring(l->connection, buffer);
	  d->last_sent_rtt = rtt;
	}
      }
    }
  }
}
  
/*---------------------------------------------------------------------------*/

struct permlink *update_permlinks(name, cp, isperm)
char *name;
struct connection *cp;
int isperm;
{
struct destination *d;
struct permlink *p;
int pl;

	if (cp != NULLCONNECTION)
		notify_destinations (NULLDESTINATION, name, cp->rev, 1);
	for (pl = 0; pl < NR_PERMLINKS; pl++)	{
		p = permarray[pl];
		if (p && !strcasecmp(p->name, name))	{
			for (d = destinations; d; d = d->next)	{
				if (!isperm && (d->link == p) && cp == NULLCONNECTION)
					notify_destinations (d, d->name, d->rev, 0);
				if (d->rtt && (d->link == p))
					update_destinations(p, d->name, 0, "");
			}
			if (cp) {
				p->connection = cp;
				p->statetime = currtime;
				p->tries = 0;
				p->waittime = 9;
				p->rxtime = 0;
				p->txtime = 0;
				p->testwaittime = currtime;
				p->testnexttime = currtime + 60;
				p->retrytime = currtime + p->waittime;
				if (isperm)
					p->permanent = isperm;
			} else {
				if (!isperm && cp == NULLCONNECTION)	{
					logit (LEV_LINKS, "disconnect from %s", p->name);
					notify_destinations (NULLDESTINATION, p->name, p->connection->rev, 0);
				}
				if (!p->permanent) {
					free(p);
					permarray[pl] = NULLPERMLINK;
				}
			}
			return p;
		}
	}
	for (pl = 0; pl < NR_PERMLINKS; pl++)	{
		if (!permarray[pl])
			break;
	}
	if (pl < NR_PERMLINKS)	{
		p = (struct permlink *) calloc(1, sizeof(*p));
		strncpy(p->name, name, HOSTNAMESIZE);
		*p->socket = '\0';
		*p->command = '\0';
		p->permanent = 0;
		p->connection = cp;
		p->statetime = currtime;
		p->tries = 0;
		p->waittime = 9;
		p->rxtime = 0;
		p->txtime = 0;
		p->testnexttime = currtime + 60;
		p->testwaittime = currtime;
		p->retrytime = currtime + p->waittime;
		p->permanent = isperm;
		permarray[pl] = p;
	} else {
		bye_command(cp);		/* no space left in array */
		return NULLPERMLINK;
	}
	return p;
}

/*---------------------------------------------------------------------------*/

void connect_permlinks()
{

#define MAX_WAITTIME   (60*60*3)

  char buffer[2048];
  int addrlen;
  int fd;
  int flags;
  int pl;
  struct connection *cp;
  struct permlink *p, *p2;
  struct sockaddr *addr;
  struct destination *d;
  int linuxbug;
#if defined(__OS2__)
  int mask = 1;
#endif
#ifdef linux
  struct sockaddr_in *sin = (struct sockaddr_in *)&addr;
#endif

  /* Update our existing connections */
  for (pl = 0; pl < NR_PERMLINKS; pl++) {
    p = permarray[pl];
    if (p && p->connection) {
      if (p->testnexttime < currtime) {
	if ((p->testwaittime + 7300) < currtime) {
	  p->rxtime = 0;
	  p->txtime = 0;
	  for (d = destinations; d; d = d->next) {
	    if (d->link == p) {
	      update_destinations(p, d->name, 0, "");
	    }
	  }
	}
	logit (LEV_HOSTCMDS, "CMD sent: /..PING");
        appendstring(p->connection, "/\377\200PING\n");
        p->testwaittime = currtime;
        p->testnexttime = currtime + 7300;
      }
    }
  }

  /* Now attempt to connect our down connections */
  for (pl = 0; pl < NR_PERMLINKS; pl++) {
    p = permarray[pl];
    if (p && !p->connection) {
      if (p->retrytime < currtime) {
       /* First see if there are any primary links. If there are and
        * they are connected, don't connect this link.
        */
       for (p2= p->primary; p2; p2=p2->primary)	{
        if (p2->connection)
		break;
       }
       if (p2 && p2->connection)	{
	 p->retrytime = currtime;
	 p->waittime = 9;
         continue;
       }

       /* Ok, no primary connections, let's try and connect */

        p->tries++;
        if (p->waittime == 9) p->waittime = 300;	/* *2 = 10 minutes */
        p->waittime <<= 1;		/* 10,20,40,80,160,180,180,... min */
        if (p->waittime > MAX_WAITTIME) p->waittime = MAX_WAITTIME;
        p->retrytime = currtime + p->waittime;

        if (!(addr = build_sockaddr(p->socket, &addrlen))) continue;
#ifdef linux
        linuxbug = 0;
        sin = (struct sockaddr_in *)&addr;
        if (sin->sin_family == AF_INET) linuxbug = 1;
#endif
        if ((fd = socket(addr->sa_family, SOCK_STREAM, 0)) < 0) continue;
#if defined(__OS2__)
        if (ioctl(fd, FIONBIO, (unsigned char *)&mask, sizeof(mask)) == -1) {
#else
#ifdef linux
        if (linuxbug) {
#endif
	if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
	  close(fd);
	  continue;
	}
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
#endif
	  logit(LEV_LINKS, "fcntl failed to %s connect, reason %d", p->name, errno);
#if defined (__OS2__)
          soclose(fd);
#else
	  close(fd);
#endif
	  continue;
	}
#ifdef linux
        }
#endif
        if (connect(fd, addr, addrlen)) {
#if defined(__OS2__)
          if (sock_errno() != SOCEINPROGRESS) {
#ifdef LOGGING
             sock_errno("connect");
#endif
             soclose(fd);
             continue;
          }
#else
	  if (errno != EINPROGRESS) {
	    logit (LEV_LINKS, "connect failed to %s, reason %d", p->name, errno);
	    close(fd);
	    continue;
	  }
#endif
        }

	cp = alloc_connection(fd, linuxbug);
	if (!cp)
		logit(LEV_WARNING, "cp == 0 on %s connect", p->name);
	p->connection = cp;
        if (*p->command) appendstring(cp, p->command);

	logit (LEV_LINKS, "connect to %s", p->name);
        /* We have made a connection. Now see if there are any backup
         * links, and if so, close them down.
         */
        for (p2= p->backup; p2; p2=p2->backup)
	        if (p2->connection)
			bye_command2(p2->connection, "backup link closed");

        appendstring(cp, convcmd);
        sprintf(buffer, "/\377\200HOST %s %s %s\n", myhostname, myrev, myfeatures);
	logit (LEV_HOSTCMDS, "CMD sent: /..%s", &buffer[3]);
        appendstring(cp, buffer);
        p->testnexttime = currtime + 5;
      }
    }
  }
}

/*---------------------------------------------------------------------------*/

static void process_input(cp)
struct connection *cp;
{
char *arg;
char buffer[2048];
int arglen;
struct cmdtable *cmdp;
struct channel *ch;
char buf[(NAMESIZE * 2) + 2];

	clear_locks();
	cp->locked = 1;

	if ((cp->type == CT_USER) && (strncmp (cp->ibuf, "/\377\200", 3)))
		convert(cp->charset_in, ISO, cp->ibuf);

	if (*cp->ibuf == '/') {
		if (cp->ibuf[1] == '\377' && cp->ibuf[2] == '\200')
			logit (LEV_HOSTCMDS, "CMD received /..%s", &cp->ibuf[3]);
		arglen = strlen(arg = getarg(cp->ibuf + 1, 0));
		for (cmdp = cmdtable; cmdp->name; cmdp++)
			if (!strncmp(cmdp->name, arg, arglen)) {
				if (cp->observer && !(cmdp->states & CM_OBSERVER))	{
					appendstring(cp, OBSERVERSTR);
					appendstring(cp, OBSERVERSTR2);
					appendprompt(cp, 0);
					return;
				}
				if (!cp->observer || cmdp->states & CM_OBSERVER)
					if (cmdp->states & (1 << cp->type)) (*cmdp->fnc)(cp);
						return;
			}
		/* didn't find match in table - see if it is a host command */
		if (cp->ibuf[1] == '\377' && cp->ibuf[2] == '\200')	{
			*(--getargp) = ' ';
			h_unknown_command(cp);	/* yep, pass it on */
			return;
		}

		if (cp->type == CT_USER) {
			if (!ecmd_exists(arg))	{
				sprintf(buffer, "*** Unknown command '/%s'. Type /HELP for help.\n", arg);
				appendstring(cp, buffer);
			} else	{
				char *params;
				params = getarg(0, 1);
				sprintf (buffer, "/\377\200ECMD %s %s %s\n", cp->name, arg, params);
				strcpy (cp->ibuf, buffer);
				h_unknown_command (cp);
			}
			appendprompt(cp, 0);
		}
		return;
	}

	if (cp->observer)
		appendstring(cp, OBSERVERSTR);
	else if (cp->type == CT_USER) {
		getTXname(cp, buf);
		if (*cp->away)
			appendstring(cp, "*** You are away, aren't you ? :-)\n");
		if (cp->expected) {
			getarg(cp->ibuf, 1);	/* just fill that static variable */
			operator_command(cp);
			cp->expected = 0;
		} else {
			if (cp->query[0] == '\0') {
				for (ch = channels; ch; ch=ch->next) {
					if (ch->chan == cp->channel)
						break;
				}
				if (cp->operator || cp->channelop || !(ch->flags & M_CHAN_M))
					send_msg_to_channel(buf, cp->channel, cp->ibuf);
				else
					appendstring(cp, "*** This is a moderated channel. Only channel operators may write.\n");
			} else
				send_msg_to_user(buf, cp->query, cp->ibuf);
		}
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void logit (level, fmt)
int level;
const char *fmt;
{
#ifdef LOGGING
va_list args;
int sysloglevel;
#ifdef __OS2__
char *cp;
#endif

	sysloglevel = (level == LEV_WARNING) ? LOG_WARNING : LOG_INFO;
	if (currentLOGlevel >= level)	{
		va_start(args,fmt);
#ifdef __OS2__
		cp = ctime(&currtime);
		cp[19] = 0;
		printf ("%s: ", &cp[4]);
		vprintf(fmt, args);
		printf ("\n");
#else
		openlog(convtype, LOG_PID, LOG_MAIL);
		vsyslog(sysloglevel, fmt, args);
		closelog();
#endif
		va_end(args);
	}
#endif
}

/*---------------------------------------------------------------------------*/

char *
strupr(s)
    char *s;
{
    register char *p = s;

    while (*p)
        *p = toupper(*p), p++;
    return s;
}

/*---------------------------------------------------------------------------*/

char *
strlwr(s)
    char *s;
{
    register char *p = s;

    while (*p)
        *p = tolower(*p), p++;
    return s;
}

/*---------------------------------------------------------------------------*/

static int
ours (buf, search)
const char *buf, *search;
{
char buf2[82], search2[82];

	if (strchr (buf, '*'))
		return ((search == NULLCHAR) ? 1 : 0);
	if (search == NULLCHAR)
		return 1;
	strcpy (buf2, buf);
	strcpy (&search2[1], search);
	search2[0] = ':';
	strlwr (buf2);
	strlwr (search2);
	return ((int)strstr (buf2, search2));
}

/*---------------------------------------------------------------------------*/

int
dohelper(name, search)
char *name;
const char *search;
{
char buf[82], buf2[100];
int count;
FILE *fp;
int foundit = 0;
int didheader = 0;

	clear_locks();
	if ((fp = fopen (helpfile, "r")) != (FILE *)0)	{
		if (search == NULLCHAR)
			didheader = 1;
		count = 0;
		do	{
			if (fgets (buf, 80, fp))	{
				if (*buf == '#')
					continue;	/* skip comments */
				if (*buf != ':')	{
					if (count)	{
						count--;
						rip (buf);
						if (!didheader)	{
							send_msg_to_user ("conversd", name, "*** Usage:");
							clear_locks();
							didheader = 1;
						}
						sprintf (buf2, "%s%s", (*buf != '/') ? "  " : "", buf);
						send_msg_to_user ("conversd", name, buf2);
						clear_locks();
					}
				} else	{	/* start of field or continuation */
					count = (ours (buf, search)) ? 32000 : 0;
					if (search == NULLCHAR && count)
						count = 1;
					if (buf[1] == ':' && count == 1)
						count--;
					if (count)
						foundit = 1;
				}
			}
		} while (!feof (fp));
		fclose (fp);
	}
	return (foundit);
}

/*---------------------------------------------------------------------------*/

static char *keywords[] = {
	"Server", "Email", "Sysinfo", "Mode",
	"Secretnum", "Security", "Link", "MaxLinks", "LogLevel", NULLCHAR};

#define NUMKEYWORDS 9


static void read_configuration()
{
FILE *fp;
char *host_name, *sock_name;
char line[1024], *cp;
int group, quality, k;
char *primary_link;
struct permlink *p, *p2;

	amprnet = 1;
	secretnumber = 0;
	restrictedmode = 0;
	if (mysysinfo)	{
		free (mysysinfo);
		mysysinfo = NULLCHAR;
	}
	if (myemailaddr)	{
		free (mysysinfo);
		mysysinfo = NULLCHAR;
	}
	if (myhostname)	{
		free(myhostname);
		myhostname = NULLCHAR;
	}
	if (!(fp = fopen(conffile, "r")))
		return;
	while (fgets(line, sizeof(line), fp)) {
		if ((line[0] == '#') || (line[0] == ';'))
			continue;
		for (k = 0; k < NUMKEYWORDS; k++)	{
			if (!strncasecmp (line, keywords[k], strlen(keywords[k])))	{
				rip (line);
				cp = getarg(line, 0);
				switch (k)	{
					case 0:	/* Server command - required */
						cp = getargcs (0, 1);
						myhostname = (char *) malloc(strlen(cp)+1);
						strcpy(myhostname, cp);
						logit (LEV_CONFIG, "Server = %s", myhostname);
						break;
					case 1:	/* Email command - required */
						cp = getargcs (0, 1);
						myemailaddr = (char *) malloc(strlen(cp)+1);
						strcpy(myemailaddr, cp);
						logit (LEV_CONFIG, "Email = %s", myemailaddr);
						break;
					case 2:	/* Sysinfo command - optional */
						cp = getargcs (0, 1);
						mysysinfo = (char *) malloc(strlen(cp)+1);
						strcpy(mysysinfo, cp);
						logit (LEV_CONFIG, "Sysinfo = %s", mysysinfo);
						break;
					case 3:	/* Mode command - optional */
						cp = getarg (0, 0);
						if (*cp == 'i')
							amprnet = 0;
						logit (LEV_CONFIG, "Mode = %s", (amprnet) ? "amprnet" : "internet");
						break;
					case 4: /* Secretnum command - optional */
						cp = getarg (0, 0);
						secretnumber = atoi(cp);
						logit (LEV_CONFIG, "Secretnum = %d", secretnumber);
						break;
					case 5:	/* Security command - optional */
						cp = getarg (0, 0);
						if (*cp == 'r')
							restrictedmode = 1;
						if (*cp == 'o')
							restrictedmode = 2;
						logit (LEV_CONFIG, "Security = %s", (!restrictedmode) ? "open" : (restrictedmode == 1) ? "restricted" : "observer");
						break;
					case 6:	/* Link command - optional (in theory) */
						if (permarray == NULLPERMARRAY)
							permarray = calloc(NR_PERMLINKS, sizeof(struct permlink *));
						logit (LEV_CONFIG, "Link = %s", getargp);
						host_name = getargcs(0, 0);
						group = atoi(getarg(0, 0));
						quality = atoi(getarg(0, 0));
						sock_name = getarg(0, 0);
						primary_link = getargcs(0, 1);
						if (*host_name && *sock_name) {
							p = update_permlinks(host_name, NULLCONNECTION, 1);
							if (p) {
								strncpy(p->name, host_name, HOSTNAMESIZE);
								strncpy(p->socket, sock_name, HOSTNAMESIZE);
								strncpy(p->command, primary_link, PCMDSIZE);
								p->group = group;
								p->quality = quality;
								p->primary= NULL;
								p->backup= NULL;
								/* If we have a primary link, find it and link it in */
								if (strncmp(sock_name, "unix:", 5) && *primary_link!='\0') {
									p->waittime = 30;
									p2= find_permlink(primary_link);
									p->primary= p2;
									p2->backup= p;
								}
							}
						}
						break;
					case 7: /* MaxLinks command - optional */
						{
						struct permlink **tmp;
						int links = NR_PERMLINKS;

							NR_PERMLINKS = atoi(getarg(0, 0));
							tmp = permarray;
							permarray = calloc(NR_PERMLINKS, sizeof(struct permlink *));
							if (tmp != NULLPERMARRAY)	{
								memcpy (permarray, tmp, links * sizeof(struct permlink *));
								free (tmp);
							}
						}
						logit (LEV_CONFIG, "MaxLinks = %d", NR_PERMLINKS);
						break;
					case 8: /* LogLevel command - optional */
						cp = getarg(0, 0);
						if (!*cp)	{
							logit (LEV_ALWAYS, "LogLevel Invalid!");
							break;
						}
						if (isdigit (*cp) || *cp == '-')
							currentLOGlevel = atoi(cp);
						else	{
							switch (*cp)	{
							case 'a':	currentLOGlevel = LEV_ALWAYS;
									break;
							case 'n':	currentLOGlevel = LEV_NOTHING;
									break;
							case 'w':	currentLOGlevel = LEV_WARNING;
									break;
							case 'l':	currentLOGlevel = LEV_LINKS;
									break;
							case 'u':	currentLOGlevel = LEV_USERS;
									break;
							case 'h':	currentLOGlevel = LEV_HOSTCMDS;
									break;
							case 'c':	if (cp[1] == 'm')
										currentLOGlevel = LEV_CMDS;
									else
										currentLOGlevel = LEV_CONFIG;
									break;
							default:	logit (LEV_ALWAYS, "LogLevel Invalid!");
									continue;

							}
						}
						logit (LEV_CONFIG, "LogLevel = %d", currentLOGlevel);
						break;
					default:
						logit (LEV_ALWAYS, "Unknown config command = %s %s", cp, getargp);
						break;	/* ignore the unknown */
				}
			}
		}
	}						
	fclose(fp);
	if (!myhostname)	{
		logit (LEV_ALWAYS, "Unknown Server name");
		exit (0);
	}
	if (!myemailaddr)	{
		logit (LEV_ALWAYS, "Unknown Email address");
		exit (0);
	}
}

/*---------------------------------------------------------------------------*/
#ifdef CHECKFILES
static void check_files_changed()
{

  static long conftime;
  static long nexttime;
  static long progtime;
  struct stat statbuf;

  if (nexttime > currtime) return;
  nexttime = currtime + 600;

  if (!stat(progfile, &statbuf)) {
    if (!progtime) progtime = statbuf.st_mtime;
    if (progtime != statbuf.st_mtime && statbuf.st_mtime < currtime - 60)
      exit(0);
  }
  if (!stat(conffile, &statbuf)) {
    if (!conftime) conftime = statbuf.st_mtime;
    if (conftime != statbuf.st_mtime && statbuf.st_mtime < currtime - 60)
      exit(0);
  }
}
#endif


char *
buildfnames (sep, ctype)
const char *sep, *ctype;
{
char *prefix;
int port;
char *pEnv = (char *) 0;
char *dir;
static char thelistener[8];
static char theconvcmd[12];

#if defined (__OS2__)
  	pEnv = getenv("ETC");
#endif
	dir = (pEnv) ? pEnv : CONF_DIR;
	pEnv = (pEnv) ? pEnv : "";
	if (!strcmp(convtype, "suconversd"))	{
		prefix = "su";
		port = 6811;
	} else if (!strcmp(convtype, "kaconversd"))	{
		prefix = "ka";
		port = 6809;
	} else if (!strcmp(convtype, "lconversd"))	{
		prefix = "l";
		port = 6810;
	} else if (!strcmp(convtype, "wconversd"))	{
		prefix = "w";
		port = 3610;
	} else {
		prefix = "";
		port = 3600;
	}
	sprintf (conffile, "%s%s%sconvers.conf", dir, sep, prefix);
	sprintf (motdfile, "%s%s%sconvers.motd", dir, sep, prefix);
	sprintf (restrictfile, "%s%s%sconvers.restrict", dir, sep, prefix);
	sprintf (accessfile, "%s%s%sconvers.access", dir, sep, prefix);
	sprintf (noaccessfile, "%s%s%sconvers.noaccess", dir, sep, prefix);
	sprintf (helpfile, "%s%s%sconvers.help", dir, sep, prefix);
	sprintf (observerfile, "%s%s%sconvers.observer", dir, sep, prefix);
	sprintf (issuefile, "%s%s%sconvers.issue", dir, sep, prefix);
	sprintf (progfile, "%s%s%sconversd", BIN_DIR, sep, prefix);
#if defined (__OS2__)
	sprintf (userfile, "%s\\spool\\convers\\log\\%sconvers.log", pEnv, prefix);
#else
	sprintf (userfile, "%s/log/%sconvers.log", DATA_DIR, prefix);
#endif
	sprintf (thelistener, "*:%-d", port);
	sprintf (listener, "unix:%s/sockets/%sconvers", DATA_DIR, prefix);
	sprintf (theconvcmd, "%sconvers\n", prefix);
	convcmd = theconvcmd;
#if 0
	printf ("conffile: '%s'\n", conffile);
	printf ("motdfile: '%s'\n", motdfile);
	printf ("restrictfile: '%s'\n", restrictfile);
	printf ("accessfile: '%s'\n", accessfile);
	printf ("noaccessfile: '%s'\n", noaccessfile);
	printf ("helpfile: '%s'\n", helpfile);
	printf ("observerfile: '%s'\n", observerfile);
	printf ("issuefile: '%s'\n", issuefile);
	printf ("progfile: '%s'\n", progfile);
	printf ("userfile: '%s'\n", userfile);
	printf ("listener: '%s'\n", listener);
	printf ("listeners[1]: '%s'\n", thelistener);
	printf ("convcmd: '%s'\n", convcmd);
	fflush (stdout);
#endif
	return thelistener;
}


/*---------------------------------------------------------------------------*/

void
update_stats (now)
time_t now;
{
int k;

	if (now >= nexthourlystats)	{
		for (k = 59; k; k--)	{
			hourly[k].tx    = hourly[k - 1].tx;
			hourly[k].rx    = hourly[k - 1].rx;
			hourly[k].start = hourly[k - 1].start;
		}
		hourly[0].start = nexthourlystats;
		nexthourlystats += ONE_HOUR;
		hoursonline++;
		hourly[0].rx = hourly[0].tx = 0;
	}
	if (now >= nextdailystats)	{
		for (k = 59; k; k--)	{
			daily[k].tx    = daily[k - 1].tx;
			daily[k].rx    = daily[k - 1].rx;
			daily[k].start = daily[k - 1].start;
		}
		daily[0].start = nextdailystats;
		nextdailystats += ONE_DAY;
		daysonline++;
		daily[0].rx = daily[0].tx = 0;
	}
}

/*---------------------------------------------------------------------------*/

void
init_stats(void)
{
struct tm *t;

	starttime = time ((time_t *)0);
	t = localtime (&starttime);
	t->tm_min = t->tm_sec = 0;
	hourly[0].start = mktime (t);
	nexthourlystats = hourly[0].start + ONE_HOUR;
	if (t->tm_hour)
		t->tm_hour = 0;
	daily[0].start = mktime(t);
	nextdailystats = daily[0].start + ONE_DAY;
	daily[0].rx = daily[0].tx = hourly[0].rx = hourly[0].tx = 0;
}


/*---------------------------------------------------------------------------*/

#if !defined(__OS2__)

void
mysighup (int sig)
{
struct connection *p;

	/* bring down all the links, users and hosts */
	for (p = connections; p; p = p->next)
		bye_command2(p, "link shutdown");

	/* now actually do the disconnects */
	free_closed_connections();
	if (sig == SIGHUP)
		logit (LEV_ALWAYS, "SIGHUP received - restarting process");
	signal (SIGHUP, mysighup);
}

/*---------------------------------------------------------------------------*/

void
loglevel (int sig)
{
	switch (sig)	{
		case SIGUSR1:	currentLOGlevel++;
				logit (LEV_ALWAYS, "Log level increase by SIGUSR1 to %d", currentLOGlevel);
				break;
		case SIGUSR2:	if (currentLOGlevel)	{
					currentLOGlevel--;
					logit (LEV_ALWAYS, "Log level decrease by SIGUSR2 to %d", currentLOGlevel);
				} else
					logit (LEV_ALWAYS, "Log level attempted decrease by SIGUSR2 to below zero");
	}
	signal (sig, loglevel);
}
#endif

/*---------------------------------------------------------------------------*/

int main(argc, argv)
int argc;
char **argv;
{
  static struct {
    char *name;
    int fd;
  } listeners[] = {
    {NULL, -1},
    {"*:3600", -1},
    {0, -1}
  };

  FD_SET_TYPE actread;
  FD_SET_TYPE actwrite;
  char *sp;
  char buffer[2048];
  int addrlen;
  int arg;
  int i;
  int size;
  int status;
  struct connection *cp;
  struct mbuf *bp;
  struct sockaddr *addr;
  struct timeval sel_timeout;
#if defined (__OS2__)
#define SEPARATOR "\\"
  char *pEnv;
#else
#define SEPARATOR "/"
#endif

#if defined (__OS2__)
  sock_init();
#else
  switch (fork())	{
	case 0:		/* child remains running */
			break;
	case -1:	/* error */
			fprintf (stderr, "Error forking Converse Server!\n");
			exit (errno);
			break;
	default:	/* parent - dies a quiet death */
			exit (0);
  }
#endif
  time (&boottime);
  chdir(SEPARATOR);
  umask(000);
#if !defined(__OS2__)
  for (i = 0; i < FD_SETSIZE; i++) close(i);
  setsid();
  signal(SIGPIPE, SIG_IGN);
  signal(SIGHUP, mysighup);
  signal(SIGUSR1, loglevel);
  signal(SIGUSR2, loglevel);
  if (!getenv("TZ")) putenv("TZ=localtime");
#endif

  sprintf(myrev, "TPP-%4.4s", strchr(REV, ':')+2);
  sp = ctime (&boottime);
  rip (sp);
  logit (LEV_CONFIG, "%s started at %s", myrev, sp);
  
  gethostname(buffer, sizeof(buffer));
  if ((sp = strchr(buffer, '.')) != NULL) *sp = 0;
  myhostname = (char *) malloc(strlen(buffer)+1);
  strcpy(myhostname, buffer);

#if defined(__OS2__)
  strcpy(convtype, "conversd");           /* for now only one conversd type supported */
#else
  sp = strrchr(argv[0],'/');
  if (sp) {
    sp++;
  } else {
    sp = argv[0];
  }
  strcpy(convtype, sp);
#endif

  listeners[1].name = buildfnames (SEPARATOR, convtype);
  listeners[0].name = listener;
  time(&currtime);

#ifdef oldLOG
  clear_logfile();
#endif
  if (argc < 2) {
    read_configuration();
  } else {
    listeners[0].name = argv[1];
    listeners[1].name = 0;
    if (argc > 2)	{
        if (myhostname) free(myhostname);
        myhostname = (char *) malloc(strlen(argv[2])+1);
        strcpy(myhostname, argv[2]);
	if (argc > 3)	{
		if (tolower(argv[3][0]) == 'r')
			restrictedmode = 1;
		if (tolower(argv[3][0]) == 'o')
			restrictedmode = 2;
	}
    }
  }

  init_stats();
  
  for (i = 0; listeners[i].name; i++)
    if ((addr = build_sockaddr(listeners[i].name, &addrlen)) &&
	(listeners[i].fd = socket(addr->sa_family, SOCK_STREAM, 0)) >= 0) {
      switch (addr->sa_family) {
      case AF_UNIX:
	unlink(addr->sa_data);
	break;
      case AF_INET:
	arg = 1;
	setsockopt(listeners[i].fd, SOL_SOCKET, SO_REUSEADDR, (char *) &arg, sizeof(arg));
	break;
      }
      if (bind(listeners[i].fd, addr, addrlen) || listen(listeners[i].fd, SOMAXCONN)) {
#if defined(__OS2__)
	soclose(listeners[i].fd);
#else
	close(listeners[i].fd);
#endif
	listeners[i].fd = -1;
      } else {
	if (maxfd < listeners[i].fd) maxfd = listeners[i].fd;
	FD_SET(listeners[i].fd, &chkread);
      }
    }

  charset_init();

  for (; ; ) {

    free_closed_connections();

#if !defined(__OS2__)
    while (waitpid(-1, &status, WNOHANG) > 0) ;
#endif
#ifdef CHECKFILES
    check_files_changed();
#endif
    connect_permlinks();

    actread = chkread;
    actwrite = chkwrite;
    sel_timeout.tv_sec = 60;
    sel_timeout.tv_usec = 0;
#ifdef POSIX
    fflush(NULL);
#endif
    if (select(maxfd + 1, &actread, &actwrite, (fd_set *) 0, &sel_timeout) <= 0) {
#if defined(__OS2__)
      psock_errno("select:");
#endif
      FD_ZERO(&actread);
      FD_ZERO(&actwrite);
    }

    if (maxfd <0) sleep(1);

    time(&currtime);

    update_stats(currtime);

    for (i = 0; listeners[i].name; i++)
      if (listeners[i].fd >= 0 && FD_ISSET(listeners[i].fd, &actread))
	accept_connect_request(listeners[i].fd);

    for (cp = connections; cp; cp = cp->next) {

      if (cp->fd >= 0 && FD_ISSET(cp->fd, &actread)) {
#if defined(__OS2__)
	size = recv(cp->fd, buffer, sizeof(buffer), 0);
#else
	size = read(cp->fd, buffer, sizeof(buffer));
#endif
	if (size <= 0) {
#if defined(__OS2__)
          psock_errno("recv:");
#endif
	  logit ("link failure with %s", cp->name);
	  bye_command2(cp, "link failure");
	} else {
	  cp->received += size;
	  daily[0].rx += size;
	  hourly[0].rx += size;
	  for (i = 0; i < size; i++)
	    switch (buffer[i]) {
	    case '\n':
 	    case '\r':
	      if (cp->icnt) {
		cp->ibuf[cp->icnt] = '\0';
		process_input(cp);
		cp->icnt = 0;
	      } else {
                appendprompt(cp, 0);
	      }
	      break;
	    case '\377':
		if (cp->icnt > 2)	{	/* no <CR> or <LF> on prev line */
			cp->ibuf[cp->icnt - 1] = '\0';
			process_input(cp);
			cp->icnt = 1;
			cp->ibuf[0] = '/';
		}
		/* and fall through */
	    default:
	      if (cp->icnt < sizeof(cp->ibuf) - 5)
		cp->ibuf[cp->icnt++] = buffer[i];
	      break;
	    }
	}
      }

      if (cp->fd >= 0 && FD_ISSET(cp->fd, &actwrite)) {
	bp = cp->obuf;
#if defined(__OS2__)
        size = send(cp->fd, bp->data, strlen(bp->data), 0);
#else
	size = write(cp->fd, bp->data, strlen(bp->data));
#endif
	if (size <= 0) {
#if defined(__OS2__)
          psock_errno("send:");
#endif
	  logit ("link failure with %s", cp->name);
	  bye_command2(cp, "link failure");
	} else {
	  cp->xmitted += size;
	  daily[0].tx += size;
	  hourly[0].tx += size;
	  bp->data += size;
	  if (!*bp->data) {
	    cp->obuf = bp->next;
	    free(bp);
	  }
	  if (!cp->obuf) FD_CLR(cp->fd, &chkwrite);
	}
      }
    }
  }
}
